import threading
from nanolab.src.snippet_manager import SnippetManager
from nanolab import pycmd
from .base_mixin import CommandMixinBase
from collections import defaultdict


class ThreadedCommandMixin(CommandMixinBase):

    def _execute_command_sequence(self, commands):
        for command_config in commands:
            self.command_instance.execute_another_command(command_config)

    def validate(self):
        for python_command in self.command_instance.command_config["commands"]:
            if "method" not in python_command:
                raise ValueError(f"Python command: 'method' is required.")

            method_name = python_command["method"]

            if "class" in python_command:
                cls = getattr(pycmd, python_command["class"])
                if not cls:
                    raise ValueError(
                        f"Python command: Class '{python_command['class']}' not found."
                    )
                if not hasattr(cls, method_name):
                    raise ValueError(
                        f"Python command: Method '{method_name}' not found in class '{python_command['class']}'."
                    )
            else:
                if method_name not in globals():
                    raise ValueError(
                        f"Python command: Method '{method_name}' not found in the current module."
                    )

    def execute(self):
        threads = []
        command_groups = defaultdict(list)

        for command_config in self.command_instance.command_config["commands"]:
            group = command_config.get("group")
            if group:
                command_groups[group].append(command_config)
            else:
                thread = threading.Thread(
                    target=self.command_instance.execute_another_command,
                    args=(command_config, ))
                threads.append(thread)

        #execute these commands in sequence in one thread
        for group_commands in command_groups.values():
            thread = threading.Thread(target=self._execute_command_sequence,
                                      args=(group_commands, ))
            threads.append(thread)

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()



import json
import ssl
import threading
import websocket
import asyncio
import time
from typing import List, Any
from nanomock.modules.nl_rpc import NanoRpc


class StatsLogger:

    def __init__(self,
                 logger_type: str,
                 node_name: str,
                 node_version: str,
                 hashes: List[Any],
                 start_block_count: dict,
                 timeout: int = 600,
                 expected_block_count=None,
                 ws_url: str = None,
                 rpc_url: str = None):

        self.logger_type = logger_type
        self.timeout = timeout
        self.node_name = node_name
        self.node_version = node_version
        self.ws_url = ws_url
        self.rpc_url = rpc_url
        self.hashes = hashes
        #self.start_block_count = start_block_count
        self.count_start = int(start_block_count["count"])
        self.cemented_start = int(start_block_count["cemented"])
        self.expected_blocks_count = expected_block_count or len(hashes)
        self.count_total = self.expected_blocks_count + self.count_start
        self.nanorpc = NanoRpc(self.rpc_url)
        self.logging_task = None

    def is_fully_synced(self):
        block_count = self.nanorpc.block_count()

        cemented = int(block_count["cemented"])
        count = int(block_count["count"])

        cemented_diff = cemented - self.cemented_start
        is_synced = cemented_diff == self.expected_blocks_count
        return is_synced, count, cemented

    def print_rpc_stats(self, elapsed_time, check_count, cemented_count,
                        previous_count, previous_cemented):
        bps = 0 if previous_count == 0 else check_count - previous_count
        cps = 0 if previous_cemented == 0 else cemented_count - previous_cemented
        print(
            f"{elapsed_time:>4} seconds {self.node_name:>12} | {self.node_version} | \
{cemented_count:>7}/{check_count:>7}/{self.count_total:>7} @{bps:>5} bps \
(avg: {round((check_count - self.count_start)/max(1, elapsed_time),2)}) | \
@{cps:>5} cps (avg: {round((cemented_count - self.cemented_start)/max(1, elapsed_time),2)})"
        )

    def print_ws_stats(self, elapsed_time, cemented_count, previous_cemented):
        cps = 0 if previous_cemented == 0 else cemented_count - previous_cemented
        print(
            f"""{elapsed_time:>4} seconds {self.node_name:>12} | {self.node_version} | @{cps:>5} cps (avg: {round((cemented_count)/max(1, elapsed_time),2)})"""
        )

    async def log_rpc(self):
        if not self.rpc_url:
            raise ValueError("rpc_url must be defined")

        previous_count = 0
        previous_cemented = 0
        start_time = time.time()

        while True:
            is_synced, check_count, cemented_count = self.is_fully_synced()
            elapsed_time = int(time.time() - start_time)
            self.print_rpc_stats(elapsed_time, check_count, cemented_count,
                                 previous_count, previous_cemented)

            if is_synced or elapsed_time > self.timeout:
                break

            previous_count = check_count
            previous_cemented = cemented_count
            await asyncio.sleep(1)

    async def log_websocket(self):
        if not self.ws_url:
            raise ValueError("ws_url must be defined")

        simple_ws = SimpleWs(self.ws_url,
                             self.node_name,
                             self.node_version,
                             ws_topics=["confirmation"])

        cemented_count = 0
        previous_cemented = 0
        start_time = time.time()

        while cemented_count < len(self.hashes):
            cemented_count = simple_ws.confirmed_hashes
            elapsed_time = int(time.time() - start_time)
            self.print_ws_stats(elapsed_time, cemented_count,
                                previous_cemented)

            if elapsed_time > self.timeout:
                break

            previous_cemented = cemented_count
            await asyncio.sleep(1)

        simple_ws.ws.close()

    async def start(self):
        if self.logger_type == "rpc":
            await self.log_rpc()
        elif self.logger_type == "ws":
            await self.log_websocket()
        else:
            raise ValueError(
                "Invalid logger_type. Must be 'rpc' or 'websocket'.")

    async def end(self):
        if self.logging_task:
            try:
                await asyncio.wait_for(self.logging_task, timeout=self.timeout)
            except asyncio.TimeoutError:
                print("Logging task timed out.")
            self.logging_task = None

    async def start_logging(self):
        await self.start()

    async def end_logging(self):
        await self.end()


class SimpleWs:

    def __init__(self,
                 ws_url: str,
                 node_name: str,
                 node_version: str,
                 ws_topics=["confirmation"]):
        self.ws_url = ws_url
        self.node_name = node_name
        self.ws_topics = ws_topics
        self.confirmed_hashes = 0
        self.lock = threading.Lock()

        self.ws = websocket.WebSocketApp(
            ws_url,
            on_message=lambda ws, msg: self.on_message(ws, msg),
            on_error=lambda ws, msg: self.on_error(ws, msg),
            on_close=lambda ws, msg, err: self.on_close(ws, msg, err),
            on_open=lambda ws: self.on_open(ws))

        if ws_url.startswith("wss"):
            self.wst = threading.Thread(
                target=self.ws.run_forever,
                kwargs={"sslopt": {
                    "cert_reqs": ssl.CERT_NONE
                }},
                daemon=True)
        else:
            self.wst = threading.Thread(target=self.ws.run_forever,
                                        daemon=True)

        self.wst.start()

    def increment_confirmed_hash_count(self):
        with self.lock:
            self.confirmed_hashes += 1

    def on_message(self, ws, message):
        self.increment_confirmed_hash_count()
        # You can implement your logic for handling WebSocket messages here.

    def on_error(self, ws, error):
        print(error)

    def on_close(self, ws, msg, error):
        print("### websocket connection closed ###")

    def on_open(self, ws):
        if "confirmation" in self.ws_topics:
            ws.send(
                json.dumps({
                    "action": "subscribe",
                    "topic": "confirmation",
                    "options": {
                        "include_election_info": "false",
                        "include_block": "true"
                    }
                }))
        print("### websocket connection opened ###")



# nanolab/command/argparse_commands.py

from os import environ
from .github import GitHubAPI
import argparse
from nanolab.src.config_handler import ConfigPathHandler, ConfigResourceHandler
from nanomock.modules.nl_parse_config import ConfigReadWrite
from nanolab.src.utils import extract_packaged_data_to_disk
from nanolab.src.config_loader import TestcaseConfig
from nanolab.src.snippet_manager import SnippetManager
from nanolab.command.command import Command


def parse_args():
    parser = argparse.ArgumentParser(
        description="Load configuration and snippets files")

    subparsers = parser.add_subparsers(dest="command")

    # Run command
    run_parser = subparsers.add_parser("run", help="Run a testcase")
    run_parser.add_argument('-t', '--testcase', type=str, help='Testcase name')
    run_parser.add_argument('-i', '--image', type=str, help='Docker image')
    run_parser.add_argument('--gh-user',
                            type=str,
                            default='gr0vity-dev',
                            help='GitHub user (default: gr0vity-dev)')
    run_parser.add_argument(
        '--gh-repo',
        type=str,
        default='nanolab-configs',
        help='GitHub repository (default: nanolab-configs)')
    run_parser.add_argument(
        '--gh-path',
        type=str,
        default='default',
        help='Path to testcases within the repository (default: default)')

    # List command
    list_parser = subparsers.add_parser("list", help="List testcases")
    list_parser.add_argument('--gh-user',
                             type=str,
                             default='gr0vity-dev',
                             help='GitHub user (default: gr0vity-dev)')
    list_parser.add_argument(
        '--gh-repo',
        type=str,
        default='nanolab-configs',
        help='GitHub repository (default: nanolab-configs)')
    list_parser.add_argument(
        '--gh-path',
        type=str,
        default='default',
        help='Path to testcases within the repository (default: default)')

    return parser.parse_args()


def _load_and_validate_configs(snippet_path, config_path):

    sinppet_manager = SnippetManager(snippet_path)
    config_loader = TestcaseConfig(sinppet_manager, config_path=config_path)

    config_loader.apply_globals()
    config_loader.complete_config()
    config_loader.validate_config()

    return config_loader


def _execute_command(command_config, snippet_manager: SnippetManager):
    if command_config.get('skip', False):
        return

    command = Command(command_config, snippet_manager)
    command.validate()
    command.execute()


def _execute_commands(config_loader: TestcaseConfig):

    for docker_tag in config_loader.config["docker_tags"]:
        environ["docker_tag"] = docker_tag
        for command_config in config_loader.config["commands"]:
            _execute_command(command_config, config_loader.snippet_manager)


class ArgParseHandler:

    def __init__(self, args=None):
        self.args = args or parse_args()
        self.conf_rw = ConfigReadWrite()
        self.github_api = GitHubAPI(self.args.gh_user, self.args.gh_repo,
                                    self.args.gh_path)

    def run(self):
        path_handler = ConfigPathHandler(self.args.testcase)
        resource_handler = ConfigResourceHandler(self.github_api, path_handler)

        #
        extract_packaged_data_to_disk(path_handler.get_snippets_path())

        #copy config to disk
        config_path = resource_handler.copy_config_file(
            path_handler.config_copy_file_path)

        #resolve and downlaod resources defiend in the config
        resolved_config = resource_handler.download_and_replace_resources(
            config_path, path_handler.downloads_path)

        self.conf_rw.write_json(path_handler.resolved_config_file_path,
                                resolved_config)

        config_loader = _load_and_validate_configs(
            path_handler.get_snippets_path(),
            path_handler.get_resolved_config_path())
        _execute_commands(config_loader)

    def list_testcases(self):
        files = self.github_api.get_files()

        if files:
            for file in files:
                if file["name"].endswith(".json"):
                    print(file["name"][:-5])

    def get_args(self):
        return self.args



from nanolab.src.snippet_manager import SnippetManager
from .base_mixin import CommandMixinBase


class SnippetCommandMixin(CommandMixinBase):

    def validate(self):
        snippet = self.command_instance.snippet_manager.get_snippet_by_key(
            self.command_instance.command_config["key"])
        missing_vars = set(snippet.get("mandatory_vars", set())) - set(
            self.command_instance.command_config.get("variables", {}).keys())
        if missing_vars:
            raise KeyError(
                f"Missing value(s) for mandatory variable(s): {', '.join(missing_vars)}"
            )

    def execute(self):
        snippet = self.command_instance.snippet_manager.get_snippet_by_key(
            self.command_instance.command_config["key"])

        for snippet_command in snippet["commands"]:
            snippet_command["variables"] = self._get_completed_variables()
            self.command_instance.execute_another_command(snippet_command)



from os import environ, path, walk
import time
import shutil
from importlib import resources
from pathlib import Path
import logging
import requests
from nanomock.modules.nl_parse_config import ConfigParser
import filecmp

logger = logging.getLogger(__name__)


def copy_if_changed(src_dir, dst_dir):
    for src_root, dirs, files in walk(src_dir):
        dst_root = src_root.replace(src_dir, dst_dir, 1)
        for file in files:
            src_file = path.join(src_root, file)
            dst_file = path.join(dst_root, file)
            if not path.exists(dst_file) or not filecmp.cmp(
                    src_file, dst_file, shallow=False):
                shutil.copy2(src_file, dst_file)


def extract_packaged_data_to_disk(snippets_path):
    with resources.path("nanolab", "__init__.py") as src_dir:
        src_dir = src_dir.parent / "snippets"
        dest_dir = snippets_path

        if src_dir.exists():
            dest_dir.mkdir(parents=True, exist_ok=True)
            copy_if_changed(str(src_dir), str(dest_dir))
            logger.info(
                "nanolab default snippets have been copied to your current working directory."
            )
        else:
            raise FileExistsError("nanolab.snippets not found.")


def get_config_parser() -> ConfigParser:
    default_path = "."
    default_file = "nl_config.toml"
    config_parser = ConfigParser(environ.get("NL_CONF_DIR", default_path),
                                 environ.get("NL_CONF_FILE", default_file))

    return config_parser


def download_data(data: str, filename: str, resource_path: Path) -> Path:
    destination = resource_path / filename.strip()

    if not destination.exists():
        print(f"Start Fetching '{filename}' to '{destination}'")
        if data.startswith("http://") or data.startswith("https://"):
            response = requests.get(data)
            with destination.open("wb") as f:
                f.write(response.content)
        else:
            with destination.open("w") as f:
                f.write(data)
        print(f"Done  Fetching '{filename}' to '{destination}'")
        wait_for_file(destination, timeout=10)
    else:
        print(f"Skip  Fetching '{filename}' to '{destination}'")

    return destination


def wait_for_file(file_path: str, timeout: int = 10):
    """Wait for a file to be written completely."""

    # Convert the timeout to seconds.
    timeout = time.time() + timeout

    # Wait for the file to exist.
    while not path.exists(file_path):
        time.sleep(0.1)
        if time.time() > timeout:
            raise TimeoutError(
                f"Timeout waiting for file {file_path} to exist.")

    # Wait for the file size to stop changing.
    old_file_size = -1
    while old_file_size != path.getsize(file_path):
        old_file_size = path.getsize(file_path)
        time.sleep(0.1)
        if time.time() > timeout:
            raise TimeoutError(
                f"Timeout waiting for file {file_path} to be written completely."
            )

    # Now the file should be written completely.



import os
import shutil
import requests
from nanomock.modules.nl_parse_config import ConfigReadWrite
from .snippet_manager import SnippetManager
from nanolab.command.command_validator import CommandValidator
from nanolab.src.utils import wait_for_file


class TestcaseConfig:

    def __init__(self,
                 snippet_manager: SnippetManager,
                 config_path: str = None):
        self.default_command_executor = "NodeInteraction"
        self.conf_rw = ConfigReadWrite()
        self.config = self._load_config(config_path)
        self.snippet_manager = snippet_manager

    def _load_resolved_path(self, config_path, env_var, default_path) -> dict:
        config_path = config_path if config_path else os.environ.get(
            env_var, default_path)

        # Wait for the file to be written completely.
        wait_for_file(config_path, timeout=1)

        return self.conf_rw.read_json(config_path)

    def _load_config(self, config_path: str) -> dict:
        return self._load_resolved_path(config_path, "CONFIG_FILE",
                                        'config.example.json')

    def apply_globals(self):
        testcase_name = self.config["testcase"]
        global_variables = self.config.get("global", {})

        # Handle downloading or copying global files
        for key, value in global_variables.items():
            global_variables[key] = self._handle_path_or_url(
                testcase_name, value)

        # Replace global variables in the commands section
        for command in self.config["commands"]:
            self._apply_variables_to_command(global_variables, command)

    def _set_default_class(self, command):
        if command.get("type") in ["python"]:
            command.setdefault("class", self.default_command_executor)

    def _handle_path_or_url(self, testcase_name: str, value: str) -> str:
        if value.startswith("http://") or value.startswith("https://"):
            destination = self._download_url(testcase_name, value)
        else:
            destination = self._copy_path(testcase_name, value)
        return destination

    def complete_config(self):
        for command in self.config["commands"]:
            self._set_default_class(command)

            if command.get("type") == "threaded":
                for python_command in command["commands"]:
                    self._set_default_class(python_command)

            if command.get("type") == "snippet":
                snippet = self.snippet_manager.get_snippet_by_key(
                    command["key"])
                for python_command in snippet["commands"]:
                    self._set_default_class(python_command)

    # Add your existing _set_default_class and _load_snippet_by_key methods here

    def validate_config(self):
        if "commands" not in self.config:
            raise ValueError(
                "The 'commands' key is missing in the config file.")

        for idx, command in enumerate(self.config["commands"]):
            command_type = command.get("type")
            if command_type not in ["bash", "snippet", "python", "threaded"]:
                raise ValueError(
                    f"Invalid command type '{command_type}' at index {idx} in 'commands'"
                )
            if command_type in ("python", "threaded"):
                CommandValidator.validate_command(
                    command, self.snippet_manager.get_snippets())
            elif command_type == "snippet":
                if "key" not in command:
                    raise ValueError(
                        f"Missing 'key' for snippet command at index {idx} in 'commands'"
                    )

    def _download_url(self, testcase_name: str, url: str) -> str:
        url = url.strip()
        filename = os.path.basename(url)
        destination = f"./resources/{testcase_name}/{filename}"
        print("DEBUG", destination, url)
        os.makedirs(os.path.dirname(destination), exist_ok=True)

        if not os.path.exists(destination):
            response = requests.get(url)
            with open(destination, "wb") as f:
                f.write(response.content)

        return destination

    def _copy_path(self, testcase_name: str, path: str) -> str:
        filename = os.path.basename(path)
        destination = f"./resources/{testcase_name}/{filename}"
        os.makedirs(os.path.dirname(destination), exist_ok=True)

        if not os.path.exists(destination):
            shutil.copy(path, destination)
        return destination

    def _apply_variables_to_command(self, global_variables: dict,
                                    command: dict):
        for key, value in command.get("variables", {}).items():
            if isinstance(value, str):
                command["variables"][key] = self._replace_global_variables(
                    global_variables, value)
            elif isinstance(value, dict):
                for sub_key, sub_value in value.items():
                    if isinstance(sub_value, str):
                        value[sub_key] = self._replace_global_variables(
                            global_variables, sub_value)

    def _replace_global_variables(self, global_variables: dict,
                                  value: str) -> str:
        for global_key, global_value in global_variables.items():
            value = value.replace(f"{{{global_key}}}", global_value)
        return value


from .command import Command


class CommandValidator:

    @staticmethod
    def validate_command(command_config: dict, snippets: dict = None):
        command = Command(command_config, snippets)
        command.validate()



import time
import nanolab.node_interaction as nni
import asyncio


class TestClass:

    def long_running_method(self, duration=1):
        time.sleep(duration)


class NodeInteraction:

    def __init__(self):
        # Your existing code...
        self.logger = None

    def publish_blocks(self,
                       publish_params,
                       logger_type=None,
                       logger_timeout=600,
                       logger_include_peers=None,
                       logger_exclude_peers=None,
                       logger_expected_count=None):
        '''
        mandatory publish_params: blocks_path, bps
        optional  publish_params: peers, split, split_skip, reverse, shuffle, 
                  start_round, end_round, subset.start_index, subset.end_index
         '''

        asyncio.run(
            nni.xnolib_publish(publish_params, logger_type, logger_timeout,
                               logger_include_peers, logger_exclude_peers,
                               logger_expected_count))
